import abc
from datetime import datetime
from functools import lru_cache
from typing import ClassVar, Dict, List, Type, Set

import numpy as np
from loguru import logger
import geopandas as gpd

from lcfm_shapefiles.labels import LCM10LabelsColors
from lcfm_shapefiles import get_filepath


class Product(abc.ABC):
    """Abstract base class for all products."""

    # Using ClassVar to indicate these are class variables, not instance variables
    nodata: ClassVar[int] = None
    data_type: ClassVar[type] = None
    product_name: ClassVar[str] = None

    # Registry to store all product classes by product_type/band_prefix
    _product_registry: ClassVar[Dict[str, Type["Product"]]] = {}

    @classmethod
    def register(cls, band_prefix: str):
        """Register a product with the given band prefix.

        Args:
            band_prefix: The prefix used by the bands of this layer

        Returns:
            A decorator function that registers the class

        Raises:
            TypeError: If the decorated class does not inherit from this class
        """
        if not cls.product_name:
            raise ValueError(f"product_name must be defined in class {cls.__name__}")
        return cls.register_product(cls.product_name, band_prefix, cls)

    @classmethod
    def register_product(
        cls, product_type: str, band_prefix: str, expected_base_class=None
    ):
        """Register a product class for a specific product type and band prefix.

        Args:
            product_type: The type of product (e.g., "LCM10", "LCM100")
            band_prefix: The prefix used by the bands of this layer
            expected_base_class: The base class that the registered class should inherit from

        Returns:
            A decorator function that registers the class

        Raises:
            TypeError: If expected_base_class is provided and the decorated class
                      does not inherit from it
        """
        # Convert product_type format for consistency (remove hyphens for registry keys)
        normalized_product_type = product_type.replace("-", "")
        key = f"{normalized_product_type}:{band_prefix}".upper()

        def decorator(subclass):
            # Check inheritance if expected_base_class is provided
            if expected_base_class and not issubclass(subclass, expected_base_class):
                raise TypeError(
                    f"{subclass.__name__} must inherit from {expected_base_class.__name__}"
                )

            cls._product_registry[key] = subclass
            return subclass

        return decorator

    @classmethod
    @abc.abstractmethod
    def create(cls, version: str, band: str, year: int = 2020) -> "Product":
        pass

    @classmethod
    def create_product(
        cls,
        product_type: str,
        version: str,
        band: str,
        year: int = 2020,
        tile_size: int | None = None,
    ) -> "Product":
        """Create a layer instance based on product type and band name.

        Args:
            product_type: The type of product (e.g., "LCM10", "LCM100")
            version: The version string
            band: The band name (will match with band prefix)
            year: The year for the product

        Returns:
            An instance of the appropriate product subclass

        Raises:
            ValueError: If no matching layer is found for the band
        """
        # Convert product_type format for consistency (remove hyphens for registry keys)
        normalized_product_type = product_type.replace("-", "").upper()
        band = band.upper()

        # Find all registered prefixes for this product type
        prefixes = [
            key.split(":", 1)[1]
            for key in cls._product_registry.keys()
            if key.startswith(f"{normalized_product_type}:")
        ]

        # First try for an exact match
        exact_key = f"{normalized_product_type}:{band}".upper()
        if exact_key in cls._product_registry:
            product_cls = cls._product_registry[exact_key]
            target_band = cls.get_target_band(band, product_cls)
            return product_cls(
                version=version, band=target_band, year=year, tile_size=tile_size
            )

        # Sort prefixes by length (descending) to match the most specific prefix first
        sorted_prefixes = sorted(prefixes, key=len, reverse=True)

        # Then try for prefix matches
        for prefix in sorted_prefixes:
            if band.startswith(prefix):
                key = f"{normalized_product_type}:{prefix}".upper()
                product_cls = cls._product_registry[key]
                target_band = cls.get_target_band(band, product_cls)
                return product_cls(
                    version=version, band=target_band, year=year, tile_size=tile_size
                )

        registered_prefixes = ", ".join(prefixes)
        raise ValueError(
            f"Unknown {normalized_product_type} band: {band}. Available band prefixes: {registered_prefixes}"
        )

    @classmethod
    def get_target_band(cls, band: str, product_cls):
        available_bands = product_cls.get_bands()
        if len(available_bands) == 1:
            target_band = available_bands[0]
        else:
            target_band = [
                b for b in available_bands if b == band.split("_")[0].upper()
            ][0]
        return target_band

    @property
    @abc.abstractmethod
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        pass

    @property
    def tile_size(self) -> int:
        """Get the tile size in degrees."""
        return self.__tile_size

    @property
    def layer_name(self):
        """Get the name of the layer."""
        return self._layer_name

    @property
    def scale(self):
        """Get the scale of the layer."""
        return self._scale

    @property
    def inverse_scale(self):
        if self._scale:
            return 1 / self._scale
        return None

    @property
    def offset(self):
        """Get the offset of the layer."""
        return self._offset

    @property
    def band(self) -> str:
        """Get the target band name for this product."""
        return self._band

    @property
    def band_names(self) -> List[str]:
        """Get the band names to set in gdal."""
        return [self._band]

    @property
    def overview_method(self):
        """Get the overview method of the layer."""
        return "MODE" if self._layer_name == "MAP" else "BILINEAR"

    @property
    def product_type(self):
        return "Categorical" if self._layer_name == "MAP" else "Continuous"

    def __init__(
        self,
        version: str,
        layer_name: str,
        band: str,
        scale: float | None = None,
        offset: float | None = None,
        year: int = 2020,
        tile_size: int | None = None,
    ):
        """Initialize Product with a version and optional year.

        Args:
            version: Version string in the format 'vXXX' where X is a digit
            year: The year for the product, defaults to 2020
            layer_name: The name of the layer
            band: The target band name of the product
            scale: The scale of the layer
            offset: The offset of the layer

        Raises:
            ValueError: If version doesn't follow the required format
        """
        # Assert that version follows the vxxx pattern
        if (
            not version.startswith("v")
            or not version[1:].isdigit()
            or (len(version) != 4)
        ):
            logger.warning("Version must start with 'v' followed by 3 digits.")

        # Set the version and year
        self.version = version
        self.year = year

        # Set the layer properties
        self._layer_name = layer_name
        self._band = band
        self._scale = scale
        self._offset = offset
        self.__tile_size = tile_size

    @classmethod
    def all_bands(cls) -> List[str]:
        """Get all bands available for this product type from all its registered subclasses.

        Returns:
            A list of all available bands for this product type
        """
        all_bands: Set[str] = set()

        # Get bands from subclasses that have get_bands method
        for product_class in cls._product_registry.values():
            if issubclass(product_class, cls):
                all_bands.update(product_class.get_bands())

        result = sorted(list(all_bands))
        return result

    @abc.abstractmethod
    def get_bands(self) -> List[str]:
        """Get all bands available for a product layer.

        Returns:
            A list of all available bands for a product layer
        """
        pass

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.

        Args:
            tile_id: The tile identifier

        Returns:
            A dictionary containing the metadata
        """
        generation_time = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")

        metadata = dict(
            copyright=(
                f"LCFM project {self.year} / Contains modified "
                f"Copernicus Sentinel data ({self.year}) processed by "
                "LCFM consortium"
            ),
            creation_time=generation_time,
            doi="undefined",
            license="CC-BY 4.0 - https://creativecommons.org/licenses/by/4.0/",
            producer="VITO NV",
            product_name=self.product_name,
            product_layer=self.layer_name,
            product_tile=tile_id,
            reference="https://land.copernicus.eu",
            time_start=f"{self.year}-01-01T00:00:00Z",
            time_end=f"{self.year}-12-31T23:59:59Z",
            version=self.version,
        )
        return metadata


class S1AnnualProduct(Product):

    product_name: ClassVar[str] = "LSF-ANNUAL-S1"
    _grid = None

    @classmethod
    def create(cls, version: str, band: str, year: int = 2020) -> "S1AnnualProduct":
        """Convenience method to create S1Annual layers."""
        return cls.create_product(cls.product_name, version, band, year)

    @property
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        return 20

    @property
    def grid(self) -> gpd.GeoDataFrame:
        if self._grid is None:
            self.set_grid()
        return self._grid

    # Note: on MEP, reading the shapefile from the package's path (get_filepath) fails
    def set_grid(self, shapefile_name=None):
        if shapefile_name is None:
            shapefile_name = get_filepath("LCFM_100p_S2-tiles_LSF-S1-2020.fgb")
        self._grid = gpd.read_file(shapefile_name)

    def _get_epsg(self, tile_id: str) -> int:
        row = self.grid[self.grid["tile"] == tile_id]
        if not row.empty:
            return int(row["epsg"].values[0])
        else:
            raise ValueError(f"Tile ID {tile_id} not found in grid.")

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.

        Args:
            tile_id: The tile identifier

        Returns:
            A dictionary containing the metadata
        """
        metadata = super().get_metadata(tile_id)

        epsg = self._get_epsg(tile_id)

        lcm_metadata = dict(
            description=f"Sentinel 1 Annual Land Surface Features at {self.resolution}m",
            platform="Sentinel-1",
            product_crs=f"EPSG:{epsg}",
            product_grid=f"MGRS tiling grid",
            sensor="SAR",
        )
        metadata.update(lcm_metadata)

        # Sort keys alphabetically
        metadata = dict(sorted(metadata.items()))

        return metadata


@S1AnnualProduct.register("GAMMA0-OBS-NB")
class S1AnnualCountProduct(S1AnnualProduct):
    """Product class for LSF-ANNUAL-S1 count product."""

    nodata: ClassVar[int] = 255
    data_type: ClassVar[type] = np.uint8

    @classmethod
    def get_bands(cls) -> List[str]:
        return ["GAMMA0-OBS-NB"]

    def __init__(
        self,
        version: str,
        band: str = "GAMMA0-OBS-NB",
        year: int = 2020,
        tile_size=None,
    ):
        super().__init__(
            version=version,
            band=band,
            year=year,
            layer_name="GAMMA0-OBS-NB",
            tile_size=tile_size,
        )


@S1AnnualProduct.register("GAMMA0-")
class S1AnnualFeaturesProduct(S1AnnualProduct):
    """Product class for LSF-ANNUAL-S1 features product."""

    nodata: ClassVar[int] = 65535
    data_type: ClassVar[type] = np.uint16

    # Define bands as a class variable
    _class_bands: ClassVar[List[str]] = None

    @classmethod
    def get_bands(cls) -> List[str]:
        if cls._class_bands is None:
            base_bands = ["VV", "VH", "VHVVRATIO"]
            bands = [
                f"GAMMA0-{band}-P{percentile}"
                for band in base_bands
                for percentile in [10, 25, 50, 75, 90]
            ]
            # Set _class_bands to the generated bands
            cls._class_bands = bands
        return cls._class_bands

    def __init__(self, version: str, band: str, year: int = 2020, tile_size=None):
        super().__init__(
            version=version,
            year=year,
            layer_name="FEATURES",
            band=band,
            scale=0.001,
            offset=45,
            tile_size=tile_size,
        )


class S1AnnualMedianFeaturesProduct(S1AnnualFeaturesProduct):
    """Product class for LSF-ANNUAL-S1 median features product."""

    # Define bands as a class variable
    _class_bands: ClassVar[List[str]] = None

    @classmethod
    def get_bands(cls) -> List[str]:
        # Select P50 bands from superclass
        if cls._class_bands is None:
            base_bands = super().get_bands()
            # Filter to only include P50 bands
            bands = [band for band in base_bands if band.endswith("-P50")]
            cls._class_bands = bands
        return cls._class_bands


class LCMProduct(Product):

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.

        Args:
            tile_id: The tile identifier

        Returns:
            A dictionary containing the metadata
        """
        metadata = super().get_metadata(tile_id)

        lcm_metadata = dict(
            description=f"Land Cover Map - {self.resolution}m",
            platform="Sentinel-1, Sentinel-2",
            product_crs="EPSG:4326",
            product_grid=f"{self.tile_size}x{self.tile_size} degree tiling grid",
            sensor="SAR, MSI",
        )
        metadata.update(lcm_metadata)

        # Sort keys alphabetically
        metadata = dict(sorted(metadata.items()))

        return metadata


class LCM10Product(LCMProduct):
    """Base class representing the Land Cover Map 10m product."""

    nodata: ClassVar[int] = LCM10LabelsColors.NO_DATA.id
    data_type: ClassVar[type] = np.uint8
    product_name: ClassVar[str] = "LCM-10"

    @classmethod
    def create(
        cls, version: str, band: str, year: int = 2020, tile_size: int = 3
    ) -> "LCM10Product":
        """Convenience method to create LCM10 layers."""
        product = cls.create_product(
            cls.product_name, version, band, year, tile_size=tile_size
        )
        return product

    @classmethod
    def source_bands(cls) -> List[str]:
        all_bands = cls.all_bands()
        all_bands.remove("MAP-CONTINUOUS")
        all_bands.append("maprgb")
        source_bands = [band.lower() for band in all_bands]
        return source_bands

    @property
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        return 10


@LCM10Product.register("MAP")
class LCM10MapProduct(LCM10Product):
    """Product class for LCM10 MAP product."""

    @classmethod
    def get_bands(cls) -> List[str]:
        return ["MAP"]

    def __init__(
        self, version: str, band: str = "MAP", year: int = 2020, tile_size: int = 3
    ):
        super().__init__(
            version=version,
            band=band,
            year=year,
            layer_name="MAP",
            tile_size=tile_size,
        )

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.

        Args:
            tile_id: The tile identifier

        Returns:
            A dictionary containing the metadata
        """
        metadata = super().get_metadata(tile_id)

        # Generate legend dynamically based on LCM10LabelsColors
        legend_items = []
        for label in LCM10LabelsColors:
            if label.id != LCM10LabelsColors.NO_DATA.id:
                # Capitalize only the first letter of the name
                name = label.name.lower().capitalize()
                legend_items.append(f"{label.id} - {name}")
        metadata["legend"] = "\n".join(legend_items)

        # Sort keys alphabetically
        metadata = dict(sorted(metadata.items()))

        return metadata


@LCM10Product.register("MAP-CONTINUOUS")
@LCM10Product.register("MAPRGB")
class LCM10MapContinuousProduct(LCM10Product):
    """Product class for LCM10 MAP-CONTINUOUS product."""

    @property
    def band_names(self) -> List[str]:
        """Get the band names to set in gdal."""
        return ["RED", "GREEN", "BLUE"]

    @classmethod
    def get_bands(cls) -> List[str]:
        return ["MAP-CONTINUOUS"]

    def __init__(
        self,
        version: str,
        band: str = "MAP-CONTINUOUS",
        year: int = 2020,
        tile_size: int = 3,
    ):
        super().__init__(
            version=version,
            year=year,
            layer_name="MAP-CONTINUOUS",
            band=band,
            tile_size=tile_size,
        )


@LCM10Product.register("PROB")
class LCM10ProbabilityProduct(LCM10Product):
    """Product class for LCM10 PROBABILITY product."""

    # Define bands as a class variable
    _class_bands: ClassVar[List[str]] = None

    @classmethod
    def get_bands(cls) -> List[str]:
        """Get all bands for this product type.
        This method lazily initializes the bands list.
        """
        if cls._class_bands is None:
            # Generate probability bands based on actual label IDs from LCM10LabelsColors
            valid_labels = [
                label
                for label in LCM10LabelsColors
                if label.id != LCM10LabelsColors.NO_DATA.id
                and label.id != LCM10LabelsColors.UNCLASSIFIABLE.id
            ]
            cls._class_bands = [f"PROB{label.id}" for label in valid_labels]
        return cls._class_bands

    def __init__(self, version: str, band: str, year: int = 2020, tile_size: int = 3):
        super().__init__(
            version=version,
            year=year,
            layer_name="PROBABILITY",
            band=band,
            scale=0.004,
            tile_size=tile_size,
        )


class LCM100Product(LCMProduct):
    """Base class representing the Land Cover Map 100m product."""

    # Using ClassVar to indicate these are class variables, not instance variables
    nodata: ClassVar[int] = 255
    data_type: ClassVar[type] = np.uint8
    product_name: ClassVar[str] = "LCM-100"

    @classmethod
    def create(cls, version: str, band: str, year: int = 2020) -> "LCM100Product":
        """Convenience method to create LCM100 layers."""
        return cls.create_product(cls.product_name, version, band, year, tile_size=30)

    @property
    def resolution(self) -> int:
        """Get the resolution of the product in meters."""
        return 100

    @property
    def tile_size(self) -> int:
        """Get the tile size in degrees."""
        return 30

    def __init__(
        self,
        version: str,
        layer_name: str,
        band: str,
        scale: float | None = None,
        offset: float | None = None,
        year: int = 2020,
        tile_size: int = 30,
    ):
        super().__init__(
            version=version,
            layer_name=layer_name,
            band=band,
            scale=scale,
            offset=offset,
            year=year,
            tile_size=tile_size,
        )
        if tile_size != 30:
            raise ValueError("LCM100Product only supports tile size of 30 degrees.")


@LCM100Product.register("MAP")
class LCM100MapProduct(LCM100Product):
    """Product class for LCM100 MAP product."""

    @classmethod
    def get_bands(cls) -> List[str]:
        return ["MAP"]

    def __init__(
        self,
        version: str,
        band: str = "MAP",
        year: int = 2020,
        tile_size: int = 30,
    ):
        super().__init__(
            version=version,
            layer_name="MAP",
            band=band,
            year=year,
            tile_size=tile_size,
        )

    def get_metadata(self, tile_id: str):
        """Generate metadata for a product.

        Args:
            tile_id: The tile identifier

        Returns:
            A dictionary containing the metadata
        """
        metadata = super().get_metadata(tile_id)
        # Generate legend dynamically based on LCM10LabelsColors
        legend_items = []
        legend_items.append("11 - Closed tree cover")
        legend_items.append("12 - Open tree cover")
        for label in LCM10LabelsColors:
            if (
                label.id != LCM10LabelsColors.NO_DATA.id
                and label.id != LCM10LabelsColors.TREE.id
            ):
                name = label.name.lower().capitalize()
                legend_items.append(f"{label.id} - {name}")
        metadata["legend"] = "\n".join(legend_items)

        # Sort keys alphabetically
        metadata = dict(sorted(metadata.items()))

        return metadata


@LCM100Product.register("CFRAC")
class LCM100CFracProduct(LCM100Product):
    """Product class for LCM100 CFRAC product."""

    # Define bands as a class variable
    _class_bands: ClassVar[List[str]] = None

    @classmethod
    def get_bands(cls) -> List[str]:
        """Get all bands for this product type.
        This method lazily initializes the bands list.
        """
        if cls._class_bands is None:
            CFRAC_NAMES = [
                "TREECOVER",
                "SHRUBLAND",
                "GRASSLAND",
                "CROPLAND",
                "BUILTUP",
                "BARE",
                "SNOW",
                "WATER",
                "WETLAND",
                "MANGROVES",
                "LICHENS",
            ]
            cls._class_bands = [f"CFRAC-{name}" for name in CFRAC_NAMES]
        return cls._class_bands

    def __init__(self, version: str, band: str, year: int = 2020, tile_size: int = 30):
        super().__init__(
            version=version,
            year=year,
            layer_name="CFRAC",
            scale=0.004,
            band=band,
            tile_size=tile_size,
        )


@LCM100Product.register("CHANGE")
class LCM100ChangeProduct(LCM100Product):
    """Product class for LCM100 CHANGE product."""

    @classmethod
    def get_bands(cls) -> List[str]:
        return ["CHANGE"]

    def __init__(
        self, version: str, band: str = "CHANGE", year: int = 2020, tile_size: int = 30
    ):
        super().__init__(
            version=version,
            year=year,
            layer_name="CHANGE",
            band=band,
            tile_size=tile_size,
        )
